from savitrigen import __version__
from savitrigen.tree import TreeClass
from savitrigen.schema import BackendSchema
from savitrigen.util import map_fields, make_ts_typehints, extract_collections
from savitrigen.template.backend import (
    ControllerTemplate,
    ModelTemplate,
    DocumentImportTemplate,
    CommonDocumentImportTemplate,
    ReferenceImportTemplate,
    IndexTemplate,
    PluginImportTemplate,
    PluginInstanceTemplate
)

@TreeClass('backend')
class BackendTree():
    def __init__(self, schema:BackendSchema):
        super().__init__()

        self._schema = schema
        self._unused_keys = [
            'translation',
            'documentation'
        ]

        self._common_collections = [
            'user',
            'file'
        ]

    def create(self):
        if not self.on_cache:
            self.copy_file('sample.env', '.env')

        self.write_file('index.ts', IndexTemplate, {
            'module_imports': self._multiline_replace(
                self._schema.plugins,
                PluginImportTemplate,
                lambda _ : {
                    'capitalized': self._capitalize(_.split('-')[-1]),
                    'plugin_collection': _
                },
            ),
            'module_instances': self._multiline_replace(
                self._schema.plugins,
                PluginInstanceTemplate,
                lambda _ : {
                    'capitalized': ' '*4 + self._capitalize(_.split('-')[-1])
                }
            )
        })

        for k, v in self._schema.collections.items():
            self.create_collection(k, v)

    def create_build_json(self):
        content = self.json_dumps({})
        self.write_file('build.json', content)

    def create_collection(self, name:str, _description:dict):
        """Creates collections

        Each module is supposed to contain 3 files:
            - index.json
            - moduleName.controller.ts
            - moduleName.model.ts

        Will also add a description to the controller file that may contain JSDoc symbols.
        """
        path = self.make_dir('collections/{}'.format(name))

        description = dict() | _description.__dict__
        description['fields'] = description.pop('_fields')

        documentation = description.get('documentation').strip()

        fields = _description.fields
        mapped_fields = map_fields(fields)

        collections = list(extract_collections(fields))
        collections_names = [ collection[0] for collection in collections ]

        for k in self._unused_keys:
            if k in description:
                del description[k]

        params = {
            'pascal_case': self._pascal_case(name),
            'camel_case': self._camel_case(name),
            'documentation': documentation,
            'type_hints': make_ts_typehints(mapped_fields),
            'savitrigen_version': __version__,
            'copyright': self._schema.meta.owner,
            **({
                'document_imports': self._multiline_replace(
                    collections_names,
                    lambda _ : CommonDocumentImportTemplate \
                        if _ in self._common_collections \
                        else DocumentImportTemplate,
                    lambda _ : {
                        'pascal_case': self._pascal_case(_),
                        'camel_case': self._camel_case(_)
                    }
                ),
                'reference_imports': self._multiline_replace(
                    collections_names,
                    lambda _ : ReferenceImportTemplate \
                        if _ not in self._common_collections \
                        else None,
                    lambda _ : {
                        'camel_case': self._camel_case(_),
                    }
                ),

            } if len(collections) > 0 else {})
        }

        with self.change_dir(path):
            self.write_file('index.json', self._json_dumps(description))
            self.write_file('{}.controller.ts'.format(name), ControllerTemplate, params)

            self.write_file('{}.model.ts'.format(name), ModelTemplate, {
                'document_imports': '// no document imports',
                'reference_imports': '// no reference imports',
            }, params)





from .controller import ControllerTemplate
from .index import IndexTemplate
from .model import ModelTemplate

from .imports import (
    DocumentImportTemplate,
    CommonDocumentImportTemplate,
    ReferenceImportTemplate,
    PluginImportTemplate,
    PluginInstanceTemplate
)



from dataclasses import dataclass

@dataclass
class FrontendSchema(object):
    meta:dict
    collections:list[str]
    default_locale:str

    plugins:list[str] = None

    notice:str = None
    signin_text:str = None

    has_releases:bool = False
    has_feedback:bool = False
    has_notification:bool = False



import json
from contextlib import suppress
from string import Template
from savitrigen.tree import TreeClass
from savitrigen.schema import FrontendSchema
from savitrigen.template.frontend import (
    IndexTemplate,
    ModuleIndexTemplate,
    RouterTemplate,
    StoreTemplate,
    PluginImportTemplate,
    PluginInstanceTemplate,
    LocaleImportTemplate,
    InternalRouterTemplate,
    InternalHomeComponentTemplate
)

@TreeClass('frontend')
class FrontendTree():
    def __init__(self, schema:FrontendSchema):
        super().__init__()
        self._schema = schema

    def create(self):
        self.create_build_json()
        self.create_module('internal', router_tpl=InternalRouterTemplate)

        translation_table = {}
        for collection_name, collection in [(k, v) for k, v in self._schema.collections.items() if v.translation]:
            for locale, translation in collection.translation.items():
                if locale not in translation_table:
                    translation_table[locale] = {}

                translation_table[locale] |= {
                    collection_name: translation
                }


        self.write_file('index.ts', IndexTemplate, {
            'module_imports': self._multiline_replace(
                self._schema.plugins,
                PluginImportTemplate,
                lambda _ : {
                    'capitalized': self._capitalize(_.split('-')[-1]),
                    'plugin_name': _
                }
            ),
            'module_instances': self._multiline_replace(
                self._schema.plugins,
                PluginInstanceTemplate,
                lambda _ : {
                    'capitalized': ' '*4 + self._capitalize(_.split('-')[-1])
                }
            ),
            'default_locale': self._schema.default_locale,
            'locale_imports': self._multiline_replace(
                translation_table.keys(),
                LocaleImportTemplate,
                lambda _ : {
                    'locale_key': ' '*6 + _,
                    'locale': _
                }
            ),
            'menu_entries': self._multiline_replace(
                self._schema.collections.keys(),
                Template(' '*8 + "'dashboard-${collection_name}',"),
                lambda _ : {
                    'collection_name': _
                }
            )
        })

        for locale, table in translation_table.items():
            self.update_i18n(locale, table)

        with self.change_dir('modules/internal'):
            self.make_dir('components/dashboard/c-home')
            self.write_file('components/dashboard/c-home/c-home.vue', InternalHomeComponentTemplate, {})


    def create_build_json(self):
        content = self._json_dumps({
            'name': '',
            'externals': {
                'variables': {
                    'productName': self._schema.meta.product.name,
                    'productLogo': 'logo.png',
                    'productLogoAlt': 'logo-alt.png',
                    'releases': self._schema.has_releases,
                    'notification': self._schema.has_notification,
                    'feedback': self._schema.has_feedback,
                }
            }
        })

        self.write_file('build.json', content)

    def create_module(self, name:str, router_tpl=None):
        path = self.make_dir('modules/{}'.format(name))

        store_path = self.make_dir('modules/{}/store'.format(name))
        components_path = self.make_dir('modules/{}/components'.format(name))

        with self.change_dir(path):
            self.write_file('index.ts', ModuleIndexTemplate, {})
            self.write_file('router.ts', router_tpl or RouterTemplate, {})
            self.write_file('store/index.ts', StoreTemplate, {})

    def update_i18n(self, locale:str, new_table):
        path = self.make_dir('i18n/{}'.format(locale))
        table = {}

        with self.change_dir(path):
            with suppress(FileNotFoundError):
                index_content = self.read_file('index.json')
                table = json.loads(index_content)

            table |= new_table
            self.write_file('index.json', self._json_dumps(table))



from string import Template

"""Used in models"""
DocumentImportTemplate = Template(
r"import { ${pascal_case}Document } from '../${camel_case}/${camel_case}.model'"
)

ReferenceImportTemplate = Template(
r"import '../${camel_case}/${camel_case}.model'"
)

CommonDocumentImportTemplate = Template(
r"import { ${pascal_case}Document } from '@savitri/backend/collections'"
)

"""Used in index"""
PluginImportTemplate = Template(
r"import { backend as ${capitalized}Module } from '$plugin_collection'"
)

PluginInstanceTemplate = Template(
r"${capitalized}Module(),"
)



from string import Template

PluginImportTemplate = Template(
r"""import { frontend as ${capitalized}Module } from '$plugin_name'"""
)

PluginInstanceTemplate = Template(
r"""${capitalized}Module(),"""
)

LocaleImportTemplate = Template(
r"""${locale_key}: require('./i18n/${locale}/index.json'),"""
)



from string import Template

ControllerTemplate = Template(
r"""import { Mutable } from '@savitri/backend'
import { ${pascal_case}Document, ${pascal_case} } from './${camel_case}.model'
import { default as Description } from './index.json'

/**
    @remarks Generated by savitrigen v${savitrigen_version}
    @copyright ${copyright}
    ${documentation}
*/

export class ${pascal_case}Controller extends Mutable<${pascal_case}Document> {
  constructor() {
    super(${pascal_case}, Description)
  }
}
""")




import pathlib
import shutil

class Cache(object):
    _path:str = './.savitricache'

    def __init__(self):
        path = pathlib.Path(self._path)

        def ignore_files(dirn, files):
            return [
                name
                for name in files
                if not pathlib.Path('{}/{}'.format(dirn, name)).is_dir()
            ]

        if not path.exists():
            shutil.copytree(
                './source',
                str(path),
                ignore=ignore_files,
                symlinks=False
            )

    def create(self, trees:list):
        for tree in trees:
            tree.set_on_cache(True)
            tree.set_silent(True)
            tree.create()



import shutil
import os
from pathlib import Path

"""Usually ~/.savitrigen"""
CONFIG_PATH = os.path.join(os.path.expanduser('~'), '.savitrigen')

"""Path that holds package files"""
SCRIPT_PATH = os.path.dirname(os.path.realpath(__file__))

def make_userdir():
    config_path = Path(CONFIG_PATH)

    if not config_path.exists():
        config_path.mkdir(exist_ok=True)

    shutil.copytree(
        os.path.join(SCRIPT_PATH, 'data'),
        CONFIG_PATH,
        dirs_exist_ok=True
    )

def copy_from_userdir(fname:str, dest:str, exist_ok:bool=False):
    if not exist_ok and Path(dest).exists():
        raise Exception('existing file')

    return shutil.copyfile(
        '{}/{}'.format(CONFIG_PATH, fname),
        dest
    )

def get_presets():
    directory = Path(CONFIG_PATH) / Path('presets')
    for f in directory.glob('*.yml'):
        yield f.name.split('.yml')[0]



from string import Template

ModuleIndexTemplate = Template(
r"""export * from './router'
import * as router from './router'
import * as store from './store'

export default () => ({
  ...router,
  ...store
})
""")

IndexTemplate = Template(
r"""import { useApp } from '@savitri/frontend'
import { SvMain } from '@savitri/components'

${module_imports}

import InternalModule from './modules/internal'

const options = {
  component: SvMain,
  modules: [
${module_instances}
    InternalModule()
  ],
  i18n: {
    locale: '${default_locale}',
    messages: {
${locale_imports}
    ...require('./i18n/pt_BR/index.json')
    }
  },
  menuSchema: {
    'Início': {
      children: [
${menu_entries}
      ]
    },
    'Schemauração': {
      children: [
        'dashboard-accessProfile',
        'dashboard-user',
      ]
    }
  }
}

useApp(options).then(({ app }) => {
  app.mount('#app')
})
""")




import pathlib
import contextlib
import logging
import typing
from shutil import copy
from string import Template
from multipledispatch import dispatch
from diff_match_patch import diff_match_patch as Diff

class PathlibWrapper(object):
    def __init__(self, silent:bool=False):
        self.silent = silent
        self.on_cache = False
        self.logger = logging.getLogger('path')
        self.parent_dir = None
        self.dmp = Diff()

    def _get_path(self, path) -> pathlib.Path:
        if isinstance(path, pathlib.Path):
            return path
        return pathlib.Path(path)

    @contextlib.contextmanager
    def change_dir(self, new_dir:str):
        previous_dir = self.parent_dir
        self.parent_dir = new_dir

        try:
            yield
        finally:
            self.parent_dir = previous_dir

    def make_dir(self, dirname:str) -> pathlib.Path:
        path = self._get_path(dirname)
        with contextlib.suppress(FileExistsError):
            path.mkdir(parents=True)
            if not self.silent:
                self.logger.info('MKDIR %s', dirname)
        return path

    def copy_file(self, src:str, dest:str) -> pathlib.Path:
        res = copy(src, '{}/{}'.format(self.parent_dir, dest))
        return self._get_path(res)

    def read_file(self, file:typing.Union[str, pathlib.Path]) -> str:
        path = self._get_path(file)
        return path.read_text()

    def _write_file(self, file:typing.Union[pathlib.PosixPath, str], content:str, patch=True) -> pathlib.Path:
        path = self._get_path(file)

        if not self.on_cache and path.exists():
            cached_path = str(path).replace('source/', '.savitricache/', 1)
            cached_path = pathlib.Path(cached_path)

            if cached_path.exists():
                cached = cached_path.read_text()
                patches = self.dmp.patch_make(cached, content)
                content, _ = self.dmp.patch_apply(patches, path.read_text())

        path.write_bytes(content.encode())
        if not self.silent:
            self.logger.info('WRITE %s', file)
        return path

    @dispatch(pathlib.PosixPath, str)
    def write_file(self, *args, **kwargs) -> pathlib.Path:
        return self._write_file(*args, **kwargs)

    @dispatch(pathlib.PosixPath, Template, dict)
    def write_file(self, fname:pathlib.PosixPath, template:Template, replacements:dict) -> pathlib.Path:
        return self._write_file(fname, template.substitute(**replacements))

    @dispatch(pathlib.PosixPath, Template, dict, dict)
    def write_file(self, fname:pathlib.PosixPath, template:Template, fallback:dict, replacements:dict) -> pathlib.Path:
        return self._write_file(fname, template.substitute(fallback, **replacements))



from string import Template

IndexTemplate = Template(
r"""import { init } from '@savitri/backend/server'

${module_imports}

const config = {
  modules: [
${module_instances}
  ]
}

init(config)
    .then(server => server.start())
"""
)



import json
import pathlib
from functools import reduce
from string import Template
from savitrigen.path import PathlibWrapper

def TreeClass(layer:str, scope:str=None):
    def decorator(cls):
        cls.layer = layer
        cls.scope = scope
        class TreeClass(cls, Tree):
            pass
        return TreeClass
    return decorator

class Tree(object):
    path = None
    parent_dir: str

    def __init__(self):
        self.path = PathlibWrapper()

    def __getattr__(self, attr):
        def wrapped_method(*args, **kwargs):
            if hasattr(self.path, attr):
                if not self.path.parent_dir:
                    self.path.parent_dir = '{}/packages/{}{}'.format(
                        'source' if not self.on_cache else '.savitricache',
                        f'{self.scope}-' if self.scope else '',
                        self.layer
                    )

                path = '{}/{}'.format(self.path.parent_dir, args[0]) \
                    if not isinstance(args[0], pathlib.Path) \
                    else args[0]

                if not isinstance(path, pathlib.Path):
                    path = pathlib.Path(path)

                return getattr(self.path, attr)(path, *args[1:], **kwargs)
        return wrapped_method

    @staticmethod
    def _multiline_replace(subjects:list, _template:Template, func) -> str:
        def reducer(a:str, item) -> str:
            result = func(item)
            template = _template(item) if callable(_template) else _template
            return a + [template.substitute(**result)] \
                if template \
                else a

        return "\n".join(reduce(reducer, subjects, []))\

    @staticmethod
    def _json_dumps(what:object) -> str:
        return json.dumps(what, indent=2, ensure_ascii=False)

    @staticmethod
    def _capitalize(string:str) -> str:
        return string[0].upper() + string[1:]

    @staticmethod
    def _pascal_case(string:str) -> str:
        return string[0].upper() + string[1:]

    @staticmethod
    def _camel_case(string:str) -> str:
        return string[0].lower() + string[1:]

    @property
    def on_cache(self) -> bool:
        return self.path.on_cache

    def set_on_cache(self, value:bool):
        self.path.on_cache = value
        self.path.parent_dir = None

    def set_silent(self, value:bool):
        self.path.silent = value



from savitrigen.tree import TreeClass
from savitrigen.schema.project import ProjectSchema

@TreeClass('..')
class ProjectTree():
    def __init__(self, schema:ProjectSchema):
        super().__init__()
        self._schema = schema

    def create(self):
        self.create_package_json()

    def create_package_json(self):
        dependencies = [
            '@savitri/common',
            '@savitri/backend',
            '@savitri/frontend',
            '@savitri/components',
            '@savitri/i18n-ptbr'
        ]

        dependencies += self._schema.plugins or []
        content = self._json_dumps({
            'dependencies': {
                dep: '*'
                for dep in dependencies
            }
        })

        self.write_file('package.json', content)



import os
import subprocess
import yaml
from pathlib import Path

from savitrigen.schema import CodegenSchema

class Bootstrap():
    @staticmethod
    def clone_repo(config:CodegenSchema):
        cwd = Path('source')
        if not Path(cwd / Path('.gitignore')).exists():
            process = subprocess.Popen([
                'git',
                'clone',
                config.boilerplate_repo_url,
                str(cwd)
            ])

            stdout, stderr = process.communicate()

        return cwd

    @staticmethod
    def install():
        if not Path(Path.cwd() / Path('source/node_modules')).exists():
            os.system('cd source && npm install')





#!/usr/bin/env python3

_REQUIRED_EXECUTABLES = [
    'git',
    'mongo',
    'npm',
    'node',
    'nodemon',
    'ts-node'
]

_MISSING_EXECUTABLE_ERROR = """The following executables are required to be accessible through $PATH in order to run this program: {}
The following one is missing: {}
"""

class MissingExecutableException(Exception):
    def __init__(self, missing:str):
        self.missing = missing
    def __str__(self):
        return _MISSING_EXECUTABLE_ERROR.format(
            ' '.join(_REQUIRED_EXECUTABLES),
            self.missing
        )


if __name__ == '__main__':
    import shutil
    import sys
    import argparse

    try:
        for n in _REQUIRED_EXECUTABLES:
            if shutil.which(n) is None:
                raise MissingExecutableException(n)
    except MissingExecutableException as e:
        print(e)
        sys.exit(1)

    parser = argparse.ArgumentParser(description='Savitrigen')
    parser.add_argument('-l', '--list-presets', action='store_true', help='lists available presets then exits')
    parser.add_argument('-p', '--preset', help='creates savitrigen.yml from a preset then exits')
    parser.add_argument('-v', '--version', action='store_true', help='displays version then exits')

    args = parser.parse_args()

    from savitrigen.commandline import main, preset, list_presets, print_version
    from savitrigen.userdir import make_userdir

    make_userdir()
    match args.preset:
        case str(): preset(args.preset)
        case None if args.list_presets: list_presets()
        case None if args.version: print_version()
        case None: main()



import yaml

import savitrigen.logging
from savitrigen.userdir import CONFIG_PATH, copy_from_userdir, get_presets
from savitrigen.tree.frontend import FrontendTree
from savitrigen.tree.backend import BackendTree
from savitrigen.tree.project import ProjectTree
from savitrigen.bootstrap import Bootstrap
from savitrigen.cache import Cache
from savitrigen.source import Source
from savitrigen.schema import (
    CodegenSchema,
    ProjectSchema,
    BackendSchema,
    FrontendSchema
)

def main():
    """Main code generation routine"""
    with open('{}/{}'.format(CONFIG_PATH, 'config.yml'), 'r') as f:
        codegen_schema = CodegenSchema(**yaml.safe_load(f))

    with open('savitrigen.yml', 'r') as f:
        project_schema = ProjectSchema(**yaml.safe_load(f))
        backend_schema = BackendSchema(
            **project_schema.backend,
            meta=project_schema.meta,
            plugins=project_schema.plugins
        )

        frontend_schema = FrontendSchema(
            **project_schema.frontend,
            meta=project_schema.meta,
            collections=backend_schema.collections,
            plugins=project_schema.plugins
        )

    Bootstrap.clone_repo(codegen_schema)

    project_tree, backend_tree, frontend_tree = (
        ProjectTree(project_schema),
        BackendTree(backend_schema),
        FrontendTree(frontend_schema)
    )

    Source().create([project_tree, backend_tree, frontend_tree])
    Cache().create([project_tree, backend_tree, frontend_tree])

    Bootstrap.install()


def preset(name:str):
    """Copies ~/.savitrigen/presets/[name].yml to current directory"""
    copy_from_userdir('presets/{}.yml'.format(name), 'savitrigen.yml')

def list_presets():
    for p in get_presets():
        print(' - {}'.format(p))

def print_version():
    from savitrigen import __version__
    print(__version__)



import typing
from dataclasses import dataclass, field
from savitrigen.guideline import (
    check_collection_naming,
    check_field_naming
)

@dataclass
class Collection(object):
    collection:str
    fields:dict
    _fields:dict = field(init=False, repr=False)

    unicon:str = ''
    strict:bool = False
    route:bool = False
    report:bool = False
    presets:list[str] = None
    methods:list[str] = None

    """Those below are unused"""
    documentation:str = 'undocumented'
    translation:dict = None

    @property
    def fields(self) -> dict:
        ms = dict()

        for k, v in self._fields.items():
            ms[k] = {
                f_k: f_v
                for f_k, f_v in v.items()
                if f_v is not None
            }

        return ms

    @fields.setter
    def fields(self, value:dict) -> None:
        self._fields = value

@dataclass
class Field(object):
    label:str
    type:str = None

    description:str = None

    required:bool = False
    readonly:bool = False
    array:bool = False
    values:list = None
    mask:str = None

    """Collection specific properties"""
    collection:str = None
    index:typing.Union[str,list[str]] = None


@dataclass
class BackendSchema(object):
    meta:dict
    collections:dict
    _collections:dict = field(init=False, repr=False)

    plugins:list[str] = None

    @property
    def collections(self) -> dict:
        return self._collections

    @collections.setter
    def collections(self, value:dict) -> None:
        ms = dict()

        for k, v in value.items():
            ms[k] = collection = Collection(**(v | dict(collection=k)))
            check_collection_naming(k)

            for f_k, f_v in collection.fields.items():
                collection.fields[f_k] = Field(**f_v).__dict__
                check_field_naming(f_k)

            if True and 'documentation' not in v:
                raise ValueError('current config requires all collections to be documented whereas "{}" is not'.format(k))

        self._collections = ms
